@@ -13,15 +13,10 @@ Important functions:
 @author: rbodo
 """
 
-from __future__ import division, absolute_import
-from __future__ import print_function, unicode_literals
-
 import os
 from importlib import import_module
 
-from future import standard_library
-
-standard_library.install_aliases()
+from snntoolbox.parsing.model_libs.keras_input_lib import load
 
 
 def run_pipeline(config, queue=None):
@@ -117,7 +112,6 @@ def run_pipeline(config, queue=None):
 
     if config.getboolean('tools', 'convert') and not is_stop(queue):
         if parsed_model is None:
-            from snntoolbox.parsing.model_libs.keras_input_lib import load
             try:
                 parsed_model = load(
                     config.get('paths', 'path_wd'),
@@ -299,7 +293,7 @@ def update_setup(config_filepath):
     os.environ['KERAS_BACKEND'] = keras_backend
     # The keras import has to happen after setting the backend environment
     # variable!
-    import keras.backend as k
+    import tensorflow.keras.backend as k
     assert k.backend() == keras_backend, \
         "Keras backend set to {} in snntoolbox config file, but has already " \
         "been set to {} by a previous keras import. Set backend " \
@@ -439,13 +433,13 @@ def update_setup(config_filepath):
 
     sample_idxs_to_test = eval(config.get('simulation', 'sample_idxs_to_test'))
     num_to_test = config.getint('simulation', 'num_to_test')
-    if not sample_idxs_to_test == []:
-        if len(sample_idxs_to_test) != num_to_test:
+    if len(sample_idxs_to_test):
+        num_required = max(sample_idxs_to_test) + 1
+        if num_required > num_to_test:
             print(dedent("""
             SNN toolbox warning: Settings mismatch. Adjusting 'num_to_test' to 
-            equal the number of 'sample_idxs_to_test'."""))
-            config.set('simulation', 'num_to_test',
-                       str(len(sample_idxs_to_test)))
+            include all 'sample_idxs_to_test'."""))
+            config.set('simulation', 'num_to_test', str(num_required))
 
     # Create log directory if it does not exist.
     if config.get('paths', 'log_dir_of_current_run') == '':
