@@ -31,10 +31,13 @@ connectivity and layer attributes:
 
 @author: rbodo
 """
+import json
+
+import pickle
 
 from abc import abstractmethod
 
-import keras
+from tensorflow import keras
 import numpy as np
 
 
@@ -504,7 +507,7 @@ class AbstractModelParser:
         previous_layers = self.get_inbound_layers(layer)
         prev_layer_output_shape = self.get_output_shape(previous_layers[0])
         if len(output_shape) < len(prev_layer_output_shape) and \
-                self.get_type(layer) != 'Flatten' and \
+                self.get_type(layer) not in {'Flatten', 'Reshape'} and \
                 self.get_type(previous_layers[0]) != 'InputLayer':
             assert len(previous_layers) == 1, \
                 "Layer to flatten must be unique."
@@ -638,10 +641,15 @@ class AbstractModelParser:
         if activation_str == 'softmax' and \
                 self.config.getboolean('conversion', 'softmax_to_relu'):
             activation = 'relu'
-            activation_str = 'relu'
             print("Replaced softmax by relu activation function.")
+        elif activation_str == 'linear' and self.get_type(layer) == 'Dense' \
+                and self.config.getboolean('conversion', 'append_softmax',
+                                           fallback=False):
+            activation = 'softmax'
+            print("Added softmax.")
+        else:
+            print("Using activation {}.".format(activation_str))
 
-        print("Using activation {}.".format(activation_str))
         attributes['activation'] = activation
 
     @abstractmethod
@@ -780,7 +788,7 @@ class AbstractModelParser:
                                                verbose=0)
         else:
             steps = int(num_to_test / batch_size)
-            score = self.parsed_model.evaluate_generator(dataflow, steps)
+            score = self.parsed_model.evaluate(dataflow, steps=steps)
         print("Top-1 accuracy: {:.2%}".format(score[1]))
         print("Top-5 accuracy: {:.2%}\n".format(score[2]))
 
@@ -907,7 +915,6 @@ def save_parameters(params, filepath, fileformat='h5'):
     """Save all layer parameters to an HDF5 file."""
 
     if fileformat == 'pkl':
-        import pickle
         pickle.dump(params, open(filepath + '.pkl', str('wb')))
     else:
         import h5py
@@ -1015,6 +1022,8 @@ def get_inbound_layers(layer):
         inbound_layers = layer._inbound_nodes[0].inbound_layers
     except AttributeError:  # For Keras backward-compatibility.
         inbound_layers = layer.inbound_nodes[0].inbound_layers
+    if not isinstance(inbound_layers, (list, tuple)):
+        inbound_layers = [inbound_layers]
     return inbound_layers
 
 
@@ -1131,7 +1140,10 @@ def get_fanout(layer, config):
     fanout = 0
     for next_layer in next_layers:
         if 'Conv' in next_layer.name and not has_stride_unity(next_layer):
-            fanout = np.zeros(layer.output_shape[1:])
+            shape = layer.output_shape
+            if 'input' in layer.name:
+                shape = fix_input_layer_shape(shape)
+            fanout = np.zeros(shape[1:])
             break
 
     for next_layer in next_layers:
@@ -1177,7 +1189,10 @@ def get_fanout_array(layer_pre, layer_post, is_depthwise_conv=False):
     sx = layer_post.strides[1]
     sy = layer_post.strides[0]
 
-    fanout = np.zeros(layer_pre.output_shape[1:])
+    shape = layer_pre.output_shape
+    if 'input' in layer_pre.name:
+        shape = fix_input_layer_shape(shape)
+    fanout = np.zeros(shape[1:])
 
     for y_pre in range(fanout.shape[0 + ax]):
         y_post = [int((y_pre + py) / sy)]
@@ -1365,7 +1380,6 @@ def get_custom_layers_dict(filepath=None):
                               'NoisySGD': NoisySGD})
 
     if filepath is not None and filepath != '':
-        import json
         with open(filepath) as f:
             kwargs = json.load(f)
             custom_layers.update(kwargs)
@@ -1407,7 +1421,6 @@ def get_custom_activations_dict(filepath=None):
         'activity_regularizer': keras.regularizers.l1}
 
     if filepath is not None and filepath != '':
-        import json
         with open(filepath) as f:
             kwargs = json.load(f)
 
@@ -1442,7 +1455,19 @@ def precision(y_true, y_pred):
     precision.
     """
 
-    import keras.backend as k
+    import tensorflow.keras.backend as k
     true_positives = k.sum(k.round(k.clip(y_true * y_pred, 0, 1)))
     predicted_positives = k.sum(k.round(k.clip(y_pred, 0, 1)))
     return true_positives / (predicted_positives + k.epsilon())
+
+
+def fix_input_layer_shape(shape):
+    """
+    tf.keras.models.load_model function introduced a bug that wraps the input
+    tensors and shapes in a single-entry list, i.e.
+    output_shape == [(None, 1, 28, 28)]. Thus we have to apply [0] here.
+    """
+
+    if len(shape) == 1:
+        return shape[0]
+    return shape
