@@ -6,15 +6,15 @@ network.
 @author: rbodo
 """
 
-from __future__ import division, absolute_import
-from __future__ import print_function, unicode_literals
-
 import os
 import numpy as np
 import matplotlib.pyplot as plt
-from future import standard_library
 
-standard_library.install_aliases()
+from snntoolbox.utils.utils import get_pearson_coefficients, extract_label, \
+    wilson_score
+from snntoolbox.conversion.utils import get_activations_batch
+from snntoolbox.simulation.utils import get_sample_activity_from_batch, \
+    spiketrains_to_rates
 
 
 def output_graphs(plot_vars, config, path=None, idx=0, data_format=None):
@@ -57,9 +57,6 @@ def output_graphs(plot_vars, config, path=None, idx=0, data_format=None):
         One of 'channels_first' or 'channels_last'.
     """
 
-    from snntoolbox.simulation.utils import spiketrains_to_rates
-    from snntoolbox.simulation.utils import get_sample_activity_from_batch
-
     if plot_vars == {}:
         return
 
@@ -152,8 +149,6 @@ def plot_layer_summaries(plot_vars, config, path=None, data_format=None):
         One of 'channels_first' or 'channels_last'.
     """
 
-    from snntoolbox.utils.utils import extract_label
-
     plot_keys = eval(config.get('output', 'plot_vars'))
 
     if len(plot_vars.keys()) == 0:
@@ -184,7 +179,7 @@ def plot_layer_summaries(plot_vars, config, path=None, data_format=None):
             plot_layer_activity(plot_vars['activations_n_l'][i],
                                 str('Activations'), newpath,
                                 data_format=data_format)
-        if 'spikerates_n_l' in plot_vars and 'activations_n_l' in plot_vars:
+        if 'spikerates' in plot_keys and 'activations_n_l' in plot_vars:
             plot_activations_minus_rates(plot_vars['activations_n_l'][i][0],
                                          plot_vars['spikerates_n_l'][i][0],
                                          name, newpath, data_format)
@@ -329,9 +324,6 @@ def plot_activations(model, x_test, path, data_format=None):
         One of 'channels_first' or 'channels_last'.
     """
 
-    from snntoolbox.conversion.utils import get_activations_batch
-    from snntoolbox.simulation.utils import get_sample_activity_from_batch
-
     activations_batch = get_activations_batch(model, x_test)
     activations = get_sample_activity_from_batch(activations_batch, 0)
     for i in range(len(activations)):
@@ -493,49 +485,6 @@ def plot_network_correlations(spikerates, layer_activations):
            fontsize=16)
 
 
-def get_pearson_coefficients(spikerates_batch, activations_batch, max_rate):
-    """
-    Compute Pearson coefficients.
-
-    Parameters
-    ----------
-
-    spikerates_batch : 
-    activations_batch :
-    max_rate: float
-        Highest spike rate.
-
-    Returns
-    -------
-    
-    co: list
-
-    """
-
-    co = []
-    for layer_num in range(len(spikerates_batch)):
-        c = []
-        for sample in range(len(spikerates_batch[0][0])):
-            s = spikerates_batch[layer_num][0][sample].flatten()
-            a = activations_batch[layer_num][0][sample].flatten()
-            if layer_num < len(spikerates_batch) - 1:
-                # Remove points at origin and saturated units, except for
-                # output layer (has too few units and gets activation of 1
-                # because of softmax).
-                ss = []
-                aa = []
-                for sss, aaa in zip(s, a):
-                    if (sss > 0 or aaa > 0) and aaa < max_rate:
-                        ss.append(sss)
-                        aa.append(aaa)
-                s = ss
-                a = aa
-            c.append(np.corrcoef(s, a)[0, 1])
-        co.append(c)
-
-    return co
-
-
 def plot_pearson_coefficients(spikerates_batch, activations_batch, config,
                               path=None):
     """
@@ -568,8 +517,6 @@ def plot_pearson_coefficients(spikerates_batch, activations_batch, config,
         Where to save the output.
     """
 
-    from snntoolbox.utils.utils import extract_label
-
     max_rate = 1. / config.getfloat('simulation', 'dt')
     co = get_pearson_coefficients(spikerates_batch, activations_batch,
                                   max_rate)
@@ -814,8 +761,6 @@ def plot_param_sweep(results, n, params, param_name, param_logscale):
         Whether to plot the parameter axis in log-scale.
     """
 
-    from snntoolbox.utils.utils import wilson_score
-
     # Compute confidence intervals of the experiments
     ci = [wilson_score(q, n) for q in results]
     if param_logscale:
@@ -1156,6 +1101,7 @@ def plot_input_image(x, label, path=None, data_format=None, filename=None):
 
     if x.ndim > 2:
         x = np.squeeze(x)
+        x = np.atleast_2d(x)
 
     plt.figure()
     plt.title('Input image (class: {})'.format(label))
@@ -1210,3 +1156,50 @@ def plot_probe(probe, path, filename):
         probe.plot()
     plt.savefig(os.path.join(path, filename))
     plt.close()
+
+
+def plot_weight_distribution(path, model):
+    parameters = model.get_weights()
+    weights = parameters[0::2]
+    biases = parameters[1::2]
+
+    plt.figure(figsize=(15, 10))
+    plt.boxplot([np.ravel(w) for w in weights], whis=15)
+    plt.xlabel("Layer index")
+    plt.ylabel("Weight value")
+    plt.savefig(os.path.join(path, 'weight_distribution'))
+
+    plt.figure(figsize=(15, 10))
+    plt.boxplot([np.ravel(b) for b in biases])
+    plt.xlabel("Layer index")
+    plt.ylabel("Bias value")
+    plt.savefig(os.path.join(path, 'bias_distribution'))
+
+
+def plot_execution_time_probe(path, probe):
+    plt.figure(figsize=(20, 5))
+    probe.plotExecutionTime()
+    plt.savefig(os.path.join(path, 'etprobe'))
+    execution_time = np.stack([probe.totalTimePerTimeStep,
+                               probe.hostTimePerTimeStep,
+                               probe.managementTimePerTimeStep,
+                               probe.learningTimePerTimeStep,
+                               probe.spikingTimePerTimeStep], -1)
+    np.savetxt(os.path.join(path, 'etprobe_csv'), execution_time, fmt='%.4e',
+               delimiter=',')
+
+
+def plot_energy_probe(path, probe):
+    plt.figure(figsize=(20, 5))
+    probe.plotEnergy()
+    plt.savefig(os.path.join(path, 'eprobe'))
+
+
+def plot_parameter_histogram(path, filename, weights, biases, bins=32):
+    plt.figure()
+    ax1 = plt.gca()
+    ax2 = ax1.twinx()
+    ax1.hist(weights.ravel(), bins=bins, label='weights', alpha=0.5)
+    ax2.hist(biases.ravel(), bins=bins, label='biases', color='m', alpha=0.5)
+    plt.legend()
+    plt.savefig(os.path.join(path, filename))
