@@ -8,18 +8,21 @@ another simulator is added to the toolbox (see :ref:`extending`).
 
 @author: rbodo
 """
-
-from __future__ import division
+import warnings
 
 import os
+import sys
 from abc import abstractmethod
-
 import numpy as np
+from tensorflow import keras
+from tensorflow.python.keras.utils.conv_utils import convert_kernel
 
-from snntoolbox.bin.utils import get_log_keys, get_plot_keys
-from snntoolbox.parsing.utils import get_type
-from snntoolbox.utils.utils import echo
-import keras
+from snntoolbox.bin.utils import get_log_keys, get_plot_keys, \
+    initialize_simulator
+from snntoolbox.conversion.utils import get_activations_batch
+from snntoolbox.parsing.utils import get_type, fix_input_layer_shape, \
+    get_fanout, get_fanin, get_outbound_layers
+from snntoolbox.utils.utils import echo, in_top_k
 
 
 class AbstractSNN:
@@ -135,7 +138,6 @@ class AbstractSNN:
     """
 
     def __init__(self, config, queue=None):
-        from snntoolbox.bin.utils import initialize_simulator
 
         self.config = config
         self.queue = queue
@@ -422,7 +424,8 @@ class AbstractSNN:
                                                               'top_k'))
 
         # Get batch input shape
-        batch_shape = list(parsed_model.layers[0].batch_input_shape)
+        batch_shape = \
+            list(fix_input_layer_shape(parsed_model.layers[0].input_shape))
         batch_shape[0] = self.batch_size
         if self.config.get('conversion', 'spike_code') == 'ttfs_dyn_thresh':
             batch_shape[0] *= 2
@@ -489,8 +492,6 @@ class AbstractSNN:
             test samples.
         """
 
-        from snntoolbox.conversion.utils import get_activations_batch
-        from snntoolbox.utils.utils import in_top_k
         if len(self._plot_keys) > 0:
             import snntoolbox.simulation.plotting as snn_plt
         else:
@@ -598,7 +599,6 @@ class AbstractSNN:
 
             # Halt if model is to be serialised only.
             if self.config.getboolean('tools', 'serialise_only'):
-                import sys
                 sys.exit()
 
             # Get classification result by comparing the guessed class (i.e.
@@ -615,8 +615,8 @@ class AbstractSNN:
             self.top1err_b_t = guesses_b_t != np.broadcast_to(
                 np.expand_dims(truth_b, -1), guesses_b_t.shape)
             for t in range(self._num_timesteps):
-                self.top5err_b_t[:, t] = ~in_top_k(output_b_l_t[:, :, t],
-                                                   truth_b, self.top_k)
+                self.top5err_b_t[:, t] = np.logical_not(
+                    in_top_k(output_b_l_t[:, :, t], truth_b, self.top_k))
 
             # Add results of current batch to previous results.
             truth_d += list(truth_b)
@@ -664,8 +664,7 @@ class AbstractSNN:
             if 'error_t' in self._plot_keys:
                 snn_plt.plot_error_vs_time(
                     self.top1err_b_t, self.top5err_b_t, self._duration,
-                    self._dt, 1 - self.top1err_ann, 1 - self.top5err_ann,
-                    log_dir)
+                    self._dt, self.top1err_ann, self.top5err_ann, log_dir)
 
             # Plot confusion matrix.
             if 'confusion_matrix' in self._plot_keys:
@@ -873,10 +872,10 @@ class AbstractSNN:
             self.input_b_l_t = np.zeros_like(self.input_b_l_t)
 
         if self.spiketrains_n_b_l_t is not None:
-            for l in range(len(self.spiketrains_n_b_l_t)):
-                self.spiketrains_n_b_l_t[l] = (
-                    np.zeros_like(self.spiketrains_n_b_l_t[l][0]),
-                    self.spiketrains_n_b_l_t[l][1])
+            for n in range(len(self.spiketrains_n_b_l_t)):
+                self.spiketrains_n_b_l_t[n] = (
+                    np.zeros_like(self.spiketrains_n_b_l_t[n][0]),
+                    self.spiketrains_n_b_l_t[n][1])
 
         if self.synaptic_operations_b_t is not None:
             self.synaptic_operations_b_t = np.zeros_like(
@@ -887,9 +886,9 @@ class AbstractSNN:
                 self.neuron_operations_b_t)
 
         if self.mem_n_b_l_t is not None:
-            for l in range(len(self.mem_n_b_l_t)):
-                self.mem_n_b_l_t[l] = (np.zeros_like(self.mem_n_b_l_t[l][0]),
-                                       self.mem_n_b_l_t[l][1])
+            for n in range(len(self.mem_n_b_l_t)):
+                self.mem_n_b_l_t[n] = (np.zeros_like(self.mem_n_b_l_t[n][0]),
+                                       self.mem_n_b_l_t[n][1])
 
     def set_connectivity(self):
         """
@@ -898,8 +897,6 @@ class AbstractSNN:
         `num_neurons`, `num_neurons_with_bias`.
         """
 
-        from snntoolbox.parsing.utils import get_fanin, get_fanout
-
         self.fanin = [0]
         self.fanout = [get_fanout(self.parsed_model.layers[0], self.config)]
         self.num_neurons = [np.product(self.parsed_model.input_shape[1:])]
@@ -977,10 +974,10 @@ class AbstractSNN:
         output_b_l_t = np.zeros(shape, 'int32')
         spiketrains_b_l_t = self.get_spiketrains_output()
         for b in range(shape[0]):
-            for l in range(shape[1]):
+            for ll in range(shape[1]):
                 for t in range(shape[2]):
-                    output_b_l_t[b, l, t] = np.count_nonzero(
-                        spiketrains_b_l_t[b, l, :t + 1])
+                    output_b_l_t[b, ll, t] = np.count_nonzero(
+                        spiketrains_b_l_t[b, ll, :t + 1])
         return output_b_l_t
 
     def reset_container_counters(self):
@@ -1158,13 +1155,14 @@ def get_samples_from_list(x_test, y_test, dataflow, config):
 
     batch_size = config.getint('simulation', 'batch_size')
     si = list(eval(config.get('simulation', 'sample_idxs_to_test')))
+    num_to_test = len(si)
     if not len(si) == 0:
         if dataflow is not None:
             batch_idx = 0
             x_test = []
             y_test = []
             target_idx = si.pop(0)
-            while len(x_test) < config.getint('simulation', 'num_to_test'):
+            while len(x_test) < num_to_test:
                 x_b_l, y_b = dataflow.next()
                 for i in range(batch_size):
                     if batch_idx * batch_size + i == target_idx:
@@ -1179,6 +1177,8 @@ def get_samples_from_list(x_test, y_test, dataflow, config):
             x_test = np.array([x_test[i] for i in si])
             y_test = np.array([y_test[i] for i in si])
 
+        config.set('simulation', 'num_to_test', str(num_to_test))
+
     return x_test, y_test
 
 
@@ -1249,11 +1249,11 @@ def build_1d_convolution(layer, delay):
                          fout * mx)
             for fin in range(weights.shape[1]):
                 source = x + (fin * nx)
-                for l in range(-px, px + 1):
-                    if not 0 <= x + l < nx:
+                for p in range(-px, px + 1):
+                    if not 0 <= x + p < nx:
                         continue
-                    connections.append((source + l, target,
-                                        weights[px - l, fin, fout], delay))
+                    connections.append((source + p, target,
+                                        weights[px - p, fin, fout], delay))
         echo('.')
     print('')
 
@@ -1288,7 +1288,6 @@ def build_convolution(layer, delay, transpose_kernel=False):
     weights, biases = get_weights(layer)
 
     if transpose_kernel:
-        from keras.utils.conv_utils import convert_kernel
         print("Transposing kernels.")
         weights = convert_kernel(weights)
 
@@ -1334,12 +1333,12 @@ def build_convolution(layer, delay, transpose_kernel=False):
                     for k in range(-py, py + 1):
                         if not 0 <= y + k < ny:
                             continue
-                        for l in range(-px, px + 1):
-                            if not 0 <= x + l < nx:
+                        for p in range(-px, px + 1):
+                            if not 0 <= x + p < nx:
                                 continue
-                            source = l + x + (y + k) * nx + fin * nx * ny
+                            source = p + x + (y + k) * nx + fin * nx * ny
                             connections.append((source, target,
-                                                weights[py - k, px - l, fin,
+                                                weights[py - k, px - p, fin,
                                                         fout], delay))
         echo('.')
     print('')
@@ -1375,7 +1374,6 @@ def build_depthwise_convolution(layer, delay, transpose_kernel=False):
     weights, biases = get_weights(layer)
 
     if transpose_kernel:
-        from keras.utils.conv_utils import convert_kernel
         print("Transposing kernels.")
         weights = convert_kernel(weights)
 
@@ -1431,12 +1429,12 @@ def build_depthwise_convolution(layer, delay, transpose_kernel=False):
                     for k in range(-py, py + 1):
                         if not 0 <= y + k < ny:
                             continue
-                        for l in range(-px, px + 1):
-                            if not 0 <= x + l < nx:
+                        for p in range(-px, px + 1):
+                            if not 0 <= x + p < nx:
                                 continue
-                            source = x + l + ((y + k) * nx) + (fin * nx * ny)
+                            source = x + p + ((y + k) * nx) + (fin * nx * ny)
                             connections.append((source, target,
-                                                weights[py - k, px - l, fin,
+                                                weights[py - k, px - p, fin,
                                                         d], delay))
             echo('.')
     print('')
@@ -1466,7 +1464,6 @@ def build_pooling(layer, delay):
     """
 
     if layer.__class__.__name__ == 'MaxPooling2D':
-        import warnings
 
         warnings.warn("Layer type 'MaxPooling' not supported yet. " +
                       "Falling back on 'AveragePooling'.", RuntimeWarning)
@@ -1492,8 +1489,8 @@ def build_pooling(layer, delay):
                              fout * nx * ny / (dx * dy))
                 for k in range(dy):
                     source = x + (y + k) * nx + fout * nx * ny
-                    for l in range(dx):
-                        connections.append((source + l, target, weight, delay))
+                    for j in range(dx):
+                        connections.append((source + j, target, weight, delay))
         echo('.')
     print('')
 
@@ -1607,8 +1604,6 @@ def get_spiking_outbound_layers(layer, config):
         List of outbound layers.
     """
 
-    from snntoolbox.parsing.utils import get_outbound_layers
-
     outbound = layer
     while True:
         outbound = get_outbound_layers(outbound)
@@ -1773,3 +1768,25 @@ def get_weights(layer):
         raise ValueError("Layer {} was expected to contain weights, biases "
                          "and, in rare cases,masks.".format(layer.name))
     return weights, biases
+
+
+def remove_name_counter(name_in):
+    """
+    Tensorflow adds a counter to layer names, e.g. <name>/kernel:0 ->
+    <name>_0/kernel:0. Need to remove this _0.
+    The situation gets complicated because SNN toolbox assigns layer names
+    that contain the layer shape, e.g. 00Conv2D_3x32x32. In addition,
+    we may get another underscore in the parameter name, e.g.
+    00DepthwiseConv2D_3X32x32_0/depthwise_kernel:0.
+    """
+
+    if '_' not in name_in or '/' not in name_in:
+        return name_in
+
+    split_dash = str(name_in).split('/')
+    assert len(split_dash) == 2, "Variable name must not contain '/'."
+    # We are only interested in the part before the /.
+    split_underscore = split_dash[0].split('_')
+    # The first '_' is assigned by SNN toolbox and should be kept.
+    return (split_underscore[0] + '_' + split_underscore[1] + '/' +
+            split_dash[1])
